#include <ESP8266WiFi.h>
#include "Adafruit_MQTT.h"
#include "Adafruit_MQTT_Client.h"
#include "DHT.h"
#include "rgb_lcd.h"

// WiFi parameters
#define WLAN_SSID "*************"                    // Insert here your WiFi network name
#define WLAN_PASS "*************"                    // Insert here the password for your WiFi network

#define AIO_SERVER      "io.adafruit.com"
#define AIO_SERVERPORT  1883
#define AIO_USERNAME    "*************"             // Insert here your AIO username
#define AIO_KEY         "*************"             // Insert here your AIO key

// DHT 11 sensor, type and digital pin
#define DHTPIN D5
#define DHTTYPE DHT11

#define soilSensorPin A0 

DHT dht(DHTPIN, DHTTYPE, 15);
rgb_lcd lcd;

int newPlantStatus = 0; // 0 indicates bad shape, 1 indicates good shape
int oldPlantStatus = 2; // 0 indicates bad shape, 1 indicates good shape
char *notifications[] = {"You should water your plant!", "Everything is cool, sit & relax :)", "Checking all measurements..."};

// Functions declarations
void connect();
WiFiClient client;

// MQTT definitions
const char MQTT_SERVER[] = AIO_SERVER;
const char MQTT_CLIENTID[] = AIO_KEY __DATE__ __TIME__;
const char MQTT_USERNAME[] = AIO_USERNAME;
const char MQTT_PASSWORD[] = AIO_KEY;

// Adafruit IO client definitions
Adafruit_MQTT_Client mqtt(&client, MQTT_SERVER, AIO_SERVERPORT, MQTT_CLIENTID, MQTT_USERNAME, MQTT_PASSWORD);

// Adafruit IO Feeds definition
const char TEMPERATURE_FEED[] = AIO_USERNAME "/feeds/Temperature";
const char HUMIDITY_FEED[] = AIO_USERNAME "/feeds/AirMoisture";
const char SOIL_FEED[] = AIO_USERNAME "/feeds/SoilMoisture";
const char ALERTS_FEED[] = AIO_USERNAME "/feeds/Alerts";

Adafruit_MQTT_Publish temperature = Adafruit_MQTT_Publish(&mqtt, TEMPERATURE_FEED);
Adafruit_MQTT_Publish humidity = Adafruit_MQTT_Publish(&mqtt, HUMIDITY_FEED);
Adafruit_MQTT_Publish soilmoisture = Adafruit_MQTT_Publish(&mqtt, SOIL_FEED);
Adafruit_MQTT_Publish alerts = Adafruit_MQTT_Publish(&mqtt, ALERTS_FEED);


void setup() {
  dht.begin();          // Initialize humidity and temperature sensor (dht)
  Serial.begin(115200);
  delay(10);
  Serial.println();

  // Connect to WiFi
  Serial.print("Connecting to ");
  Serial.println(WLAN_SSID);
  WiFi.begin(WLAN_SSID, WLAN_PASS);
  while (WiFi.status() != WL_CONNECTED) {
    delay(500);
    Serial.print(".");
    delay(10);
  }
  Serial.println("");
  Serial.println("WiFi connected");
  Serial.println("IP address: ");
  Serial.println(WiFi.localIP());

  // Initialize LCD
  lcd.begin(20, 4);
  Serial.println();

  // Connect to MQTT and Adafruit IO
  connect();

}

void loop() {
  lcd.clear();
  // ping Adafruit IO a few times to make sure we remain connected
  if (!mqtt.ping(3)) {
    if (!mqtt.connected()) connect();     // reconnect to adafruit io
  }

  // Read from sensors: humidity, temperature, soil moisture
  int humidity_data = (int)dht.readHumidity();
  int temperature_data = (int)dht.readTemperature();
  int soil_moisture_data = (int) analogRead(soilSensorPin);
  
  
  // Publish data to the relevant feeds of Adafruit IO
  if (!temperature.publish(temperature_data)) {
    Serial.println("Failed to publish temperature");
  } else {
    Serial.println("Temperature published to the feed!");
    
  } if (!humidity.publish(humidity_data)) {
    Serial.println("Failed to publish humidity");
  } else {
    Serial.println("Humidity published to the feed!");
    
  } if (!soilmoisture.publish(soil_moisture_data)) {
    Serial.println("Failed to publish soil moisture");
  } else {
    Serial.println("Soil moisture published to the feed!");
  }
  // Calculate measurements to determine the plant's status
  checkStatus(humidity_data, soil_moisture_data, temperature_data);
  // Show relevant data on the LCD
  showLcd(temperature_data, humidity_data);
  
  // Repeat every 10 seconds
  delay(10000);
}

/**
 * This method connects to the Adafruit IO and MQTT 
 **/
void connect() {
  Serial.print("Connecting to Adafruit IO... ");
  int8_t ret;
  while ((ret = mqtt.connect()) != 0) {
    switch (ret) {
      case 1: Serial.println("Wrong protocol"); break;
      case 2: Serial.println("ID rejected"); break;
      case 3: Serial.println("Server unavail"); break;
      case 4: Serial.println("Bad user/pass"); break;
      case 5: Serial.println("Not authed"); break;
      case 6: Serial.println("Failed to subscribe"); break;
      default: Serial.println("Connection failed"); break;
    }
    if (ret >= 0) {
      mqtt.disconnect();
    }
    Serial.println("Retrying connection...");
    delay(5000);
  }
  Serial.println("Adafruit IO Connected!");
}

/**
* The method calculates the measurements of the sensors to monitor the plant and
* determine whether the user has to water the plant.
* Soil sensor measurements:
* 0 to 300 sensor in water
* 300 to 700 humid soil
* 700 to 950 soil is dry
* Humidity is by percentage, temperature is by celsius
**/
void checkStatus(int humidity, int soilMoisture, int temperature) {
  // Checking all the cases where the user needs to water its plant
  if ((temperature > 22 && soilMoisture > 700 && humidity < 50) || // when it is hot and dry and the soil is dry
      (temperature < 22 && soilMoisture > 700 && humidity < 50) || // when it is cold and dry and the soil is dry
      (temperature > 22 && soilMoisture > 700 && humidity > 50)) { // when it is hot and humid and the soil is dry
          newPlantStatus = 0;
  } else {
      newPlantStatus = 1;
  }

   // Send the data to the feedonly when there is a change in the plant's shape,
   // so only in critical times the notification is sent to the user
  if (oldPlantStatus != newPlantStatus) {
    oldPlantStatus = newPlantStatus;
    alerts.publish(notifications[newPlantStatus]);
    }
}

/**
 * This method shows the relevant data in the lcd screen
 **/
void showLcd(int temperature, int humidity) {
  lcd.setCursor(0,0); 
  lcd.print("Temp: ");
  lcd.print(temperature);
  lcd.print((char)223);
  lcd.print("C");
  
  lcd.setCursor(0,1);
  lcd.print("Humidity: ");
  lcd.print(humidity);
  lcd.print("%");
}
